#include <linux/fs.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/cdev.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <asm/system.h>
#include <linux/device.h>
#include <asm/param.h>
#include <linux/timer.h>
#include <linux/delay.h>
//#include "font.h"

#define BASE_ADRESS 0x0378
#define CTRL_REG BASE_ADRESS+2
#define GLCD_ENABLE 	1	/* /strobe */
#define GLCD_RESET  	2	/* /autofeed */ 
#define GLCD_CS			4	/* init */
#define GLCD_ADRESS		8	/* /select */

/** TODO: 
 * auswahl text/grafik per ioctrl
 * ausgabe des akt. modes (text/grafik)
 * auslesen des glcd_buffer
 * positionieren mittels offset
 * gleichzeitigen zugriff blocken
 * configmöglichkeiten per ioctl (display off etc. )
 * */


dev_t glcd_dev=0;
int glcd_major=0;
int glcd_minor=0;
int glcd_nr=3; // number of ioports

static unsigned char glcd_buffer[1024];

static struct class *glcd_class;
static struct cdev *glcd_cdev=NULL;

// glcd_functions
/** set the statuslines for the handshake
 *   @param enable_ 
 *   @param reset_
 *   @param chipselect_   
 *   @param adress_       ctrl or dataregister
 *   */
void writeCtrlReg(unsigned char enable_,unsigned char adress_, unsigned char reset_, unsigned char chipselect_){
    unsigned char value=0;
    if(!enable_)    value |= GLCD_ENABLE;
    if(!reset_)     value |= GLCD_RESET;
    if(chipselect_) value |= GLCD_CS;
    if(!adress_)    value |= GLCD_ADRESS;
    outb_p(value,CTRL_REG);
//	wmb();
}


/** Write Data to GLCD
 * 
 * @param data to write
 * @param reg for choosing data or ctrl
 *  */
void writeGlcd(unsigned char data,unsigned char reg){
    outb_p(data,BASE_ADRESS);
//	wmb();
    writeCtrlReg(0,reg,1,0);
    writeCtrlReg(1,reg,1,0);
    writeCtrlReg(0,reg,1,1);

}

/** write Data to Ctrlregister
 *  @param data to write
 *   */
void writeCtrl(unsigned char data){
    writeGlcd(data,0);
}
/** write Data to Dataregister in the glcd
 *  @param data to write
 *  */
void writeData(unsigned char data){
    writeGlcd(data,1);
    
}
/** set the page in the glcd
 * @param pagenumber 
 * */
static int set_glcd_page(unsigned char pagenumber){
	if(pagenumber > 7)
		return -1;
	writeCtrl(0xb0 | pagenumber);
	return 0;
}
/** sets the position of the virtual cursor
 * ypos is only allowed from 0 to 7, and xpos from 0 to 127
 * */
static int set_glcd_cursor(unsigned char xpos,unsigned char ypos){
	unsigned tmp=xpos;
	if((xpos > 127)||(set_glcd_page(ypos)!=0)){
		printk("glcd doesnt support this size %dx%d\n",xpos,ypos);
		return -1;
		}
	tmp=(tmp>>4) & 0x0f;
	writeCtrl(0x10 | tmp); // higher nibble
	writeCtrl(xpos & 0x0f); // lower nibble
	

	return 0;
}

/** writes the buffer complete to the glcd
 * */    
static void write_glcd_buffer_to_glcd(void){
	unsigned char x,y;
	unsigned int tmp;
	for(y=0;y<8;y++){
		set_glcd_cursor(0,y);
		for(x=0;x<128;x++){
			writeData(glcd_buffer[x+(y<<7)]);		
		}
	}

}


/** initialize the lcd
* */
static int init_glcd(void){
    unsigned char array_init[]={0x40,0xa0,0xa3,0xc0,0x2f,0x20,0xac,0xA4,0xaf,0xa6}; // 10 werte
	unsigned char tmp;
	printk("init_glcd\n");
	writeCtrlReg(0,0,0,1);
	writeCtrlReg(0,0,1,1);
	for(tmp=0;tmp<10;tmp++)
		writeCtrl(array_init[tmp]);
	write_glcd_buffer_to_glcd();	
	return 0;
}

static void glcd_write_char(void){

}

static int glcd_open(struct inode * glcd_inode, struct file * glcd_filp){
// goto position 0 in display
	set_glcd_cursor(0,0);
    return 0;
}

static ssize_t glcd_write(struct file * glcd_filp, __user const char *glcd_user , size_t to_copy, loff_t * offset){
	int not_copied;
	unsigned long time;

	if(to_copy>1024){
		printk("to much data to write %ld\n",(long int)to_copy);
		return -EFBIG;
	}
	//to_copy=min(to_copy,sizeof(glcd_buffer));
	not_copied=copy_from_user(&glcd_buffer[0],glcd_user,to_copy);
	if(not_copied){
		printk("%dbytes not copied!\n",not_copied);
	}
	time=jiffies;
	write_glcd_buffer_to_glcd();	
	printk("time for print %ld\n",jiffies-time);
	msleep(100); 	
	return to_copy-not_copied;
}

static int glcd_close(struct inode * glcd_inode, struct file * glcd_filp){
    return 0;

}
static int glcd_ioctl(struct inode * glcd_inode, struct file * glcd_filp, unsigned int x, unsigned long y){
    return 0;

}
/*
static ssize_t glcd_read(struct file glcd_instance,__user char *user, size_t to_copy, loff_t *offset){

	return 0;
}*/
static struct file_operations glcd_fops = {
            .owner= THIS_MODULE,
            .write= glcd_write,
//			.read= glcd_read,
            .ioctl= glcd_ioctl,
            .open=  glcd_open,
			.release= glcd_close,
                                                         };

static int __init mod_init(void)
{
    unsigned int tmp,result;
    dev_t dev=0;
    printk("mod_init called\n");
    glcd_major=register_chrdev(glcd_major,"glcd",&glcd_fops);
    if(glcd_major<0){
        printk(KERN_WARNING "can't get Major ERROR: %d\n",glcd_major);
    	goto glcd_init_exit;
	}

	if(request_region(BASE_ADRESS,glcd_nr,"glcd")==0){
		printk("request_region_failed\n");
		goto glcd_init_unregister_chrdev;
	}
	glcd_cdev=cdev_alloc();
	if(glcd_cdev==0){
		printk("cdev_alloc failed\n");
		goto glcd_init_free_region;
	}

	glcd_class=class_create(THIS_MODULE,"glcd");
	class_device_create(glcd_class,NULL,MKDEV(glcd_major,0),NULL,"glcd%d",glcd_minor);
	if(IS_ERR(glcd_class)){
		printk("no udev support\n");
	}
	init_glcd();
    return 0;

glcd_init_free_region:
	release_region(BASE_ADRESS,glcd_nr);
	
glcd_init_unregister_chrdev:
       unregister_chrdev(glcd_major,"glcd");
glcd_init_exit:
	return -1;
}
static void __exit mod_exit(void)
{
       dev_t dev=MKDEV(glcd_major,0);//glcd_minor);
	   if(!IS_ERR(glcd_class)){
	   		class_device_destroy(glcd_class,dev);
			class_destroy(glcd_class);
		}
       printk("mod_exit called\n");
	   release_region(BASE_ADRESS,glcd_nr);
       unregister_chrdev(glcd_major,"glcd");

}

module_init(mod_init);
module_exit(mod_exit);

//###########
MODULE_AUTHOR("SiO");
MODULE_LICENSE("GPL v2");
MODULE_DESCRIPTION("simple driver for the glcd from pollin with sed1565.");
MODULE_SUPPORTED_DEVICE("pollin glcd with sed1536");
